# pylint: disable=missing-class-docstring,missing-function-docstring
# pylint: disable=line-too-long,invalid-name,attribute-defined-outside-init
"""pure python trajectory calculation backend"""

import math
from dataclasses import dataclass
from typing import NamedTuple

from .drag_model import DragDataPoint
from .conditions import Atmo, Shot, Wind
from .munition import Ammo
from .trajectory_data import TrajectoryData, TrajFlag
from .unit import Distance, Angular, Velocity, Weight, Energy, Pressure, Temperature, PreferredUnits

__all__ = (
    'TrajectoryCalc',
    'get_global_max_calc_step_size',
    'get_global_use_powder_sensitivity',
    'set_global_max_calc_step_size',
    'set_global_use_powder_sensitivity',
    'reset_globals'
)

cZeroFindingAccuracy = 0.000005
cMinimumVelocity = 50.0
cMaximumDrop = -15000
cMaxIterations = 20
cGravityConstant = -32.17405

_globalUsePowderSensitivity = False
_globalMaxCalcStepSize = Distance.Foot(0.5)


def get_global_max_calc_step_size() -> Distance:
    return _globalMaxCalcStepSize


def get_global_use_powder_sensitivity() -> bool:
    return _globalUsePowderSensitivity


def reset_globals() -> None:
    global _globalUsePowderSensitivity, _globalMaxCalcStepSize
    _globalUsePowderSensitivity = False
    _globalMaxCalcStepSize = Distance.Foot(0.5)


def set_global_max_calc_step_size(value: [float, Distance]) -> None:
    global _globalMaxCalcStepSize
    if (_value := PreferredUnits.distance(value)).raw_value <= 0:
        raise ValueError("_globalMaxCalcStepSize have to be > 0")
    _globalMaxCalcStepSize = PreferredUnits.distance(value)


def set_global_use_powder_sensitivity(value: bool) -> None:
    global _globalUsePowderSensitivity
    if not isinstance(value, bool):
        raise TypeError(f"set_global_use_powder_sensitivity {value=} is not a boolean")
    _globalUsePowderSensitivity = value


class CurvePoint(NamedTuple):
    """Coefficients for quadratic interpolation"""
    a: float
    b: float
    c: float


@dataclass
class Vector:
    x: float
    y: float
    z: float

    def magnitude(self):
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def mul_by_const(self, a: float):
        return Vector(self.x * a, self.y * a, self.z * a)

    def mul_by_vector(self, b: 'Vector'):
        return self.x * b.x + self.y * b.y + self.z * b.z

    def add(self, b: 'Vector'):
        return Vector(self.x + b.x, self.y + b.y, self.z + b.z)

    def subtract(self, b: 'Vector'):
        return Vector(self.x - b.x, self.y - b.y, self.z - b.z)

    def negate(self):
        return Vector(-self.x, -self.y, -self.z)

    def normalize(self):
        m = self.magnitude()
        if math.fabs(m) < 1e-10:
            return Vector(self.x, self.y, self.z)
        return self.mul_by_const(1.0 / m)

    def __add__(self, other: 'Vector'):
        return self.add(other)

    def __radd__(self, other: 'Vector'):
        return self.add(other)

    def __iadd__(self, other: 'Vector'):
        return self.add(other)

    def __sub__(self, other: 'Vector'):
        return self.subtract(other)

    def __rsub__(self, other: 'Vector'):
        return self.subtract(other)

    def __isub__(self, other: 'Vector'):
        return self.subtract(other)

    def __mul__(self, other: [int, float, 'Vector']):
        if isinstance(other, (int, float)):
            return self.mul_by_const(other)
        if isinstance(other, Vector):
            return self.mul_by_vector(other)
        raise TypeError(other)

    def __rmul__(self, other: [int, float, 'Vector']):
        return self.__mul__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return self.negate()


class TrajectoryCalc:
    """All calculations are done in units of feet and fps"""

    def __init__(self, ammo: Ammo):
        self.ammo = ammo
        self._bc = self.ammo.dm.BC
        self._table_data = ammo.dm.drag_table
        self._curve = calculate_curve(self._table_data)
        self.gravity_vector = Vector(.0, cGravityConstant, .0)

    @staticmethod
    def get_calc_step(step: float = 0):
        """Keep step under max_calc_step_size
        :param step: proposed step size
        :return: step size for calculations (in feet)
        """
        preferred_step = _globalMaxCalcStepSize >> Distance.Foot
        if step == 0:
            return preferred_step / 2.0
        return min(step, preferred_step) / 2.0

    def trajectory(self, shot_info: Shot, max_range: Distance, dist_step: Distance,
                   extra_data: bool = False):
        filter_flags = TrajFlag.RANGE

        if extra_data:
            dist_step = Distance.Foot(0.2)
            filter_flags = TrajFlag.ALL

        self._init_trajectory(shot_info)
        return self._trajectory(shot_info, max_range >> Distance.Foot, dist_step >> Distance.Foot, filter_flags)

    def _init_trajectory(self, shot_info: Shot):
        self.look_angle = shot_info.look_angle >> Angular.Radian
        self.twist = shot_info.weapon.twist >> Distance.Inch
        self.length = shot_info.ammo.dm.length >> Distance.Inch
        self.diameter = shot_info.ammo.dm.diameter >> Distance.Inch
        self.weight = shot_info.ammo.dm.weight >> Weight.Grain
        self.barrel_elevation = shot_info.barrel_elevation >> Angular.Radian
        self.barrel_azimuth = shot_info.barrel_azimuth >> Angular.Radian
        self.sight_height = shot_info.weapon.sight_height >> Distance.Foot
        self.cant_cosine = math.cos(shot_info.cant_angle >> Angular.Radian)
        self.cant_sine = math.sin(shot_info.cant_angle >> Angular.Radian)
        self.alt0 = shot_info.atmo.altitude >> Distance.Foot
        self.calc_step = self.get_calc_step()
        if _globalUsePowderSensitivity:
            self.muzzle_velocity = shot_info.ammo.get_velocity_for_temp(shot_info.atmo.temperature) >> Velocity.FPS
        else:
            self.muzzle_velocity = shot_info.ammo.mv >> Velocity.FPS
        self.stability_coefficient = self.calc_stability_coefficient(shot_info.atmo)

    def zero_angle(self, shot_info: Shot, distance: Distance) -> Angular:
        """Iterative algorithm to find barrel elevation needed for a particular zero
        :param shot_info: Shot parameters
        :param distance: Zero distance
        :return: Barrel elevation to hit height zero at zero distance
        """
        self._init_trajectory(shot_info)

        zero_distance = math.cos(self.look_angle) * (distance >> Distance.Foot)
        height_at_zero = math.sin(self.look_angle) * (distance >> Distance.Foot)
        maximum_range = zero_distance - 1.5 * self.calc_step
        self.barrel_azimuth = 0.0
        self.barrel_elevation = math.atan(height_at_zero / zero_distance)
        self.twist = 0

        iterations_count = 0
        zero_finding_error = cZeroFindingAccuracy * 2
        # x = horizontal distance down range, y = drop, z = windage
        while zero_finding_error > cZeroFindingAccuracy and iterations_count < cMaxIterations:
            # Check height of trajectory at the zero distance (using current self.barrel_elevation)
            t = self._trajectory(shot_info, maximum_range, zero_distance, TrajFlag.NONE)[0]
            height = t.height >> Distance.Foot
            zero_finding_error = math.fabs(height - height_at_zero)
            if zero_finding_error > cZeroFindingAccuracy:
                # Adjust barrel elevation to close height at zero distance
                self.barrel_elevation -= (height - height_at_zero) / zero_distance
            else:  # last barrel_elevation hit zero!
                break
            iterations_count += 1

        if zero_finding_error > cZeroFindingAccuracy:
            # TODO: Don't raise exception; return a tuple that contains the error so caller can check how close zero is
            raise Exception(f'Zero vertical error {zero_finding_error} feet, after {iterations_count} iterations.')
        return Angular.Radian(self.barrel_elevation)

    def _trajectory(self, shot_info: Shot, maximum_range: float, step: float,
                    filter_flags: TrajFlag) -> list[TrajectoryData]:
        """Calculate trajectory for specified shot
        :param maximum_range: Feet down range to stop calculation
        :param step: Frequency (in feet down range) to record TrajectoryData
        :return: list of TrajectoryData, one for each dist_step, out to max_range
        """
        ranges = []  # Record of TrajectoryData points to return
        ranges_length = int(maximum_range / step) + 1
        time = 0
        previous_mach = .0
        drag = 0

        # region Initialize wind-related variables to first wind reading (if any)
        len_winds = len(shot_info.winds)
        current_wind = 0
        current_item = 0
        next_range_distance = .0
        next_wind_range = Wind.MAX_DISTANCE_FEET
        if len_winds < 1:
            wind_vector = Vector(.0, .0, .0)
        else:
            wind_vector = wind_to_vector(shot_info.winds[0])
            next_wind_range = shot_info.winds[0].until_distance >> Distance.Foot
        # endregion

        # region Initialize velocity and position of projectile
        velocity = self.muzzle_velocity
        # x: downrange distance, y: drop, z: windage
        range_vector = Vector(.0, -self.cant_cosine * self.sight_height, -self.cant_sine * self.sight_height)
        velocity_vector = Vector(math.cos(self.barrel_elevation) * math.cos(self.barrel_azimuth),
                                 math.sin(self.barrel_elevation),
                                 math.cos(self.barrel_elevation) * math.sin(self.barrel_azimuth)) * velocity
        # endregion

        # With non-zero look_angle, rounding can suggest multiple adjacent zero-crossings
        seen_zero = TrajFlag.NONE  # Record when we see each zero crossing, so we only register one
        if range_vector.y >= 0:
            seen_zero |= TrajFlag.ZERO_UP  # We're starting above zero; we can only go down
        elif range_vector.y < 0 and self.barrel_elevation < self.look_angle:
            seen_zero |= TrajFlag.ZERO_DOWN  # We're below and pointing down from look angle; no zeroes!

        # region Trajectory Loop
        while range_vector.x <= maximum_range + self.calc_step:
            _flag = TrajFlag.NONE

            # Update wind reading at current point in trajectory
            if range_vector.x >= next_wind_range:
                current_wind += 1
                if current_wind >= len_winds:  # No more winds listed after this range
                    wind_vector = Vector(.0, .0, .0)
                    next_wind_range = Wind.MAX_DISTANCE_FEET
                else:
                    wind_vector = wind_to_vector(shot_info.winds[current_wind])
                    next_wind_range = shot_info.winds[current_wind].until_distance >> Distance.Foot

            # Update air density at current point in trajectory
            density_factor, mach = shot_info.atmo.get_density_factor_and_mach_for_altitude(
                self.alt0 + range_vector.y)

            # region Check whether to record TrajectoryData row at current point
            if filter_flags:
                # Zero-crossing checks
                if range_vector.x > 0:
                    # Zero reference line is the sight line defined by look_angle
                    reference_height = range_vector.x * math.tan(self.look_angle)
                    # If we haven't seen ZERO_UP, we look for that first
                    if not seen_zero & TrajFlag.ZERO_UP:
                        if range_vector.y >= reference_height:
                            _flag |= TrajFlag.ZERO_UP
                            seen_zero |= TrajFlag.ZERO_UP
                    # We've crossed above sight line; now look for crossing back through it
                    elif not seen_zero & TrajFlag.ZERO_DOWN:
                        if range_vector.y < reference_height:
                            _flag |= TrajFlag.ZERO_DOWN
                            seen_zero |= TrajFlag.ZERO_DOWN

                # Mach crossing check
                if (velocity / mach <= 1) and (previous_mach > 1):
                    _flag |= TrajFlag.MACH

                # Next range check
                if range_vector.x >= next_range_distance:
                    _flag |= TrajFlag.RANGE
                    next_range_distance += step
                    current_item += 1

                # Record TrajectoryData row
                if _flag & filter_flags:
                    ranges.append(create_trajectory_row(
                        time, range_vector, velocity_vector,
                        velocity, mach, self.spin_drift(time), self.look_angle,
                        density_factor, drag, self.weight, _flag.value
                    ))
                    if current_item == ranges_length:
                        break
            # endregion

            previous_mach = velocity / mach

            # region Ballistic calculation step (point-mass)
            # Time step is set to advance bullet calc_step distance along x axis
            delta_time = self.calc_step / velocity_vector.x
            # Air resistance seen by bullet is ground velocity minus wind velocity relative to ground
            velocity_adjusted = velocity_vector - wind_vector
            velocity = velocity_adjusted.magnitude()  # Velocity relative to air
            # Drag is a function of air density and velocity relative to the air
            drag = density_factor * velocity * self.drag_by_mach(velocity / mach)
            # Bullet velocity changes due to both drag and gravity
            velocity_vector -= (velocity_adjusted * drag - self.gravity_vector) * delta_time
            # Bullet position changes by velocity times the time step
            delta_range_vector = Vector(self.calc_step,
                                        velocity_vector.y * delta_time,
                                        velocity_vector.z * delta_time)
            # Update the bullet position
            range_vector += delta_range_vector
            velocity = velocity_vector.magnitude()  # Velocity relative to ground
            time += delta_range_vector.magnitude() / velocity

            if velocity < cMinimumVelocity or range_vector.y < cMaximumDrop:
                break
            # endregion
        # endregion
        # If filter_flags == 0 then all we want is the ending value
        if not filter_flags:
            ranges.append(create_trajectory_row(
                time, range_vector, velocity_vector,
                velocity, mach, self.spin_drift(time), self.look_angle,
                density_factor, drag, self.weight, _flag.value))
        return ranges

    def drag_by_mach(self, mach: float) -> float:
        """ Drag force = V^2 * Cd * AirDensity * S / 2m where:
                cStandardDensity of Air = 0.076474 lb/ft^3
                S is cross-section = d^2 pi/4, where d is bullet diameter in inches
                m is bullet mass in pounds
            bc contains m/d^2 in units lb/in^2, which we multiply by 144 to convert to lb/ft^2
            Thus: The magic constant found here = StandardDensity * pi / (4 * 2 * 144)
        :return: Drag coefficient at the given mach number
        """
        cd = calculate_by_curve(self._table_data, self._curve, mach)
        return cd * 2.08551e-04 / self._bc

    def spin_drift(self, time) -> float:
        """Litz spin-drift approximation
        :param time: Time of flight
        :return: windage due to spin drift, in feet
        """
        if self.twist != 0:
            sign = 1 if self.twist > 0 else -1
            return sign * (1.25 * (self.stability_coefficient + 1.2)
                           * math.pow(time, 1.83)) / 12
        return 0

    def calc_stability_coefficient(self, atmo: Atmo) -> float:
        """Miller stability coefficient"""
        if self.twist and self.length and self.diameter:
            twist_rate = math.fabs(self.twist) / self.diameter
            length = self.length / self.diameter
            # Miller stability formula
            sd = 30 * self.weight / (
                    math.pow(twist_rate, 2) * math.pow(self.diameter, 3) * length * (1 + math.pow(length, 2))
            )
            # Velocity correction factor
            fv = math.pow(self.muzzle_velocity / 2800, 1.0 / 3.0)
            # Atmospheric correction
            ft = atmo.temperature >> Temperature.Fahrenheit
            pt = atmo.pressure >> Pressure.InHg
            ftp = ((ft + 460) / (59 + 460)) * (29.92 / pt)
            return sd * fv * ftp
        return 0


def wind_to_vector(wind: Wind) -> Vector:
    """Calculate wind vector to add to projectile velocity vector each iteration:
        Aerodynamic drag is function of velocity relative to the air stream.

    Wind angle of zero is blowing from behind shooter
    Wind angle of 90-degree is blowing towards shooter's right

    NOTE: Presently we can only define Wind in the x-z plane, not any vertical component.
    """
    # Downrange (x-axis) wind velocity component:
    range_component = (wind.velocity >> Velocity.FPS) * math.cos(wind.direction_from >> Angular.Radian)
    # Cross (z-axis) wind velocity component:
    cross_component = (wind.velocity >> Velocity.FPS) * math.sin(wind.direction_from >> Angular.Radian)
    return Vector(range_component, 0, cross_component)


def create_trajectory_row(time: float, range_vector: Vector, velocity_vector: Vector,
                          velocity: float, mach: float, spin_drift: float, look_angle: float,
                          density_factor: float, drag: float, weight: float, flag: int) -> TrajectoryData:
    """
    Create a TrajectoryData object representing a single row of trajectory data.

    :param time: Time of flight.
    :param range_vector: Vector representing range.
    :param velocity_vector: Vector representing velocity.
    :param velocity: Velocity value.
    :param mach: Mach number.
    :param spin_drift: Spin drift value.
    :param look_angle: Look angle value.
    :param density_factor: Density factor.
    :param drag: Drag value.
    :param weight: Weight value.
    :param flag: Flag value.

    :return: A TrajectoryData object representing the trajectory data.
    """
    windage = range_vector.z + spin_drift
    drop_adjustment = get_correction(range_vector.x, range_vector.y)
    windage_adjustment = get_correction(range_vector.x, windage)
    trajectory_angle = math.atan(velocity_vector.y / velocity_vector.x)

    return TrajectoryData(
        time=time,
        distance=Distance.Foot(range_vector.x),
        velocity=Velocity.FPS(velocity),
        mach=velocity / mach,
        height=Distance.Foot(range_vector.y),
        target_drop=Distance.Foot((range_vector.y - range_vector.x * math.tan(look_angle)) * math.cos(look_angle)),
        drop_adj=Angular.Radian(drop_adjustment - (look_angle if range_vector.x else 0)),
        windage=Distance.Foot(windage),
        windage_adj=Angular.Radian(windage_adjustment),
        look_distance=Distance.Foot(range_vector.x / math.cos(look_angle)),
        angle=Angular.Radian(trajectory_angle),
        density_factor=density_factor - 1,
        drag=drag,
        energy=Energy.FootPound(calculate_energy(weight, velocity)),
        ogw=Weight.Pound(calculate_ogw(weight, velocity)),
        flag=flag
    )


def get_correction(distance: float, offset: float) -> float:
    """:return: Sight adjustment in radians"""
    if distance != 0:
        return math.atan(offset / distance)
    return 0  # None


def calculate_energy(bullet_weight: float, velocity: float) -> float:
    """:return: energy in ft-lbs"""
    return bullet_weight * math.pow(velocity, 2) / 450400


def calculate_ogw(bullet_weight: float, velocity: float) -> float:
    """:return: Optimal Game Weight in pounds"""
    return math.pow(bullet_weight, 2) * math.pow(velocity, 3) * 1.5e-12


def calculate_curve(data_points: list[DragDataPoint]) -> list[CurvePoint]:
    """Piecewise quadratic interpolation of drag curve
    :param data_points: List[{Mach, CD}] data_points in ascending Mach order
    :return: List[CurvePoints] to interpolate drag coefficient
    """
    # rate, x1, x2, x3, y1, y2, y3, a, b, c
    # curve = []
    # curve_point
    # num_points, len_data_points, len_data_range

    rate = (data_points[1].CD - data_points[0].CD
            ) / (data_points[1].Mach - data_points[0].Mach)
    curve = [CurvePoint(0, rate, data_points[0].CD - data_points[0].Mach * rate)]
    len_data_points = int(len(data_points))
    len_data_range = len_data_points - 1

    for i in range(1, len_data_range):
        x1 = data_points[i - 1].Mach
        x2 = data_points[i].Mach
        x3 = data_points[i + 1].Mach
        y1 = data_points[i - 1].CD
        y2 = data_points[i].CD
        y3 = data_points[i + 1].CD
        a = ((y3 - y1) * (x2 - x1) - (y2 - y1) * (x3 - x1)) / (
                (x3 * x3 - x1 * x1) * (x2 - x1) - (x2 * x2 - x1 * x1) * (x3 - x1))
        b = (y2 - y1 - a * (x2 * x2 - x1 * x1)) / (x2 - x1)
        c = y1 - (a * x1 * x1 + b * x1)
        curve_point = CurvePoint(a, b, c)
        curve.append(curve_point)

    num_points = len_data_points
    rate = (data_points[num_points - 1].CD - data_points[num_points - 2].CD) / \
           (data_points[num_points - 1].Mach - data_points[num_points - 2].Mach)
    curve_point = CurvePoint(
        0, rate, data_points[num_points - 1].CD - data_points[num_points - 2].Mach * rate
    )
    curve.append(curve_point)
    return curve


def calculate_by_curve(data: list, curve: list, mach: float) -> float:
    """
    Binary search for drag coefficient based on Mach number
    :param data: data
    :param curve: Output of calculate_curve(data)
    :param mach: Mach value for which we're searching for CD
    :return float: drag coefficient
    """
    num_points = int(len(curve))
    mlo = 0
    mhi = num_points - 2

    while mhi - mlo > 1:
        mid = int(math.floor(mhi + mlo) / 2.0)
        if data[mid].Mach < mach:
            mlo = mid
        else:
            mhi = mid

    if data[mhi].Mach - mach > mach - data[mlo].Mach:
        m = mlo
    else:
        m = mhi
    curve_m = curve[m]
    return curve_m.c + mach * (curve_m.b + curve_m.a * mach)



# pylint: disable=wildcard-import
"""Check for available backends"""

from .logger import logger

# try to use cython based backend
try:
    from py_ballisticcalc_exts import (TrajectoryCalc,
                                       get_global_max_calc_step_size,
                                       get_global_use_powder_sensitivity,
                                       set_global_max_calc_step_size,
                                       set_global_use_powder_sensitivity,
                                       reset_globals)

    logger.info("Binary modules found, running in binary mode")
except ImportError as error:
    from .trajectory_calc import (TrajectoryCalc,
                                  get_global_max_calc_step_size,
                                  get_global_use_powder_sensitivity,
                                  set_global_max_calc_step_size,
                                  set_global_use_powder_sensitivity,
                                  reset_globals)

    logger.warning("Library running in pure python mode. "
                   "For better performance install 'py_ballisticcalc.exts' package")

__all__ = (
    'TrajectoryCalc',
    'get_global_max_calc_step_size',
    'get_global_use_powder_sensitivity',
    'set_global_max_calc_step_size',
    'set_global_use_powder_sensitivity',
    'reset_globals',
)



"""
Useful types for prefer_units of measurement conversion for ballistics calculations
"""

import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass, MISSING, Field
from enum import IntEnum
from math import pi, atan, tan
from typing import NamedTuple, Union, TypeVar

from py_ballisticcalc.logger import logger


__all__ = ('Unit', 'AbstractUnit', 'AbstractUnitType', 'UnitProps',
           'UnitPropsDict', 'Distance',
           'Velocity', 'Angular', 'Temperature', 'Pressure',
           'Energy', 'Weight', 'Dimension', 'PreferredUnits')


AbstractUnitType = TypeVar('AbstractUnitType', bound='AbstractUnit')


# pylint: disable=invalid-name
class Unit(IntEnum):
    """
    Usage of IntEnum simplify data serializing for using it with databases etc.
    """
    Radian = 0
    Degree = 1
    MOA = 2
    Mil = 3
    MRad = 4
    Thousandth = 5
    InchesPer100Yd = 6
    CmPer100M = 7
    OClock = 8

    Inch = 10
    Foot = 11
    Yard = 12
    Mile = 13
    NauticalMile = 14
    Millimeter = 15
    Centimeter = 16
    Meter = 17
    Kilometer = 18
    Line = 19

    FootPound = 30
    Joule = 31

    MmHg = 40
    InHg = 41
    Bar = 42
    hPa = 43
    PSI = 44

    Fahrenheit = 50
    Celsius = 51
    Kelvin = 52
    Rankin = 53

    MPS = 60
    KMH = 61
    FPS = 62
    MPH = 63
    KT = 64

    Grain = 70
    Ounce = 71
    Gram = 72
    Pound = 73
    Kilogram = 74
    Newton = 75

    @property
    def key(self) -> str:
        """
        :return: readable name of the unit of measure
        """
        return UnitPropsDict[self].name

    @property
    def accuracy(self) -> int:
        """
        :return: default accuracy of the unit of measure
        """
        return UnitPropsDict[self].accuracy

    @property
    def symbol(self) -> str:
        """
        :return: short symbol of the unit of measure in CI
        """
        return UnitPropsDict[self].symbol

    def __repr__(self) -> str:
        return UnitPropsDict[self].name

    def __call__(self: 'Unit', value: [int, float, AbstractUnitType] = None) -> AbstractUnitType:
        """Creates new unit instance by dot syntax
        :param self: unit as Unit enum
        :param value: numeric value of the unit
        :return: AbstractUnit instance
        """

        # if value is None:
        #     return self

        if isinstance(value, AbstractUnit):
            return value << self
        if 0 <= self < 10:
            obj = Angular(value, self)
        elif 10 <= self < 20:
            obj = Distance(value, self)
        elif 30 <= self < 40:
            obj = Energy(value, self)
        elif 40 <= self < 50:
            obj = Pressure(value, self)
        elif 50 <= self < 60:
            obj = Temperature(value, self)
        elif 60 <= self < 70:
            obj = Velocity(value, self)
        elif 70 <= self < 80:
            obj = Weight(value, self)
        else:
            raise TypeError(f"{self} Unit is not supported")
        return obj


class UnitProps(NamedTuple):
    """Properties of unit measure"""
    name: str
    accuracy: int
    symbol: str


UnitPropsDict = {
    Unit.Radian: UnitProps('radian', 6, 'rad'),
    Unit.Degree: UnitProps('degree', 4, '°'),
    Unit.MOA: UnitProps('MOA', 2, 'MOA'),
    Unit.Mil: UnitProps('mil', 2, 'mil'),
    Unit.MRad: UnitProps('mrad', 2, 'mrad'),
    Unit.Thousandth: UnitProps('thousandth', 2, 'ths'),
    Unit.InchesPer100Yd: UnitProps('inch/100yd', 2, 'in/100yd'),
    Unit.CmPer100M: UnitProps('cm/100m', 2, 'cm/100m'),
    Unit.OClock: UnitProps('hour', 2, 'h'),

    Unit.Inch: UnitProps("inch", 1, "inch"),
    Unit.Foot: UnitProps("foot", 2, "ft"),
    Unit.Yard: UnitProps("yard", 1, "yd"),
    Unit.Mile: UnitProps("mile", 3, "mi"),
    Unit.NauticalMile: UnitProps("nautical mile", 3, "nm"),
    Unit.Millimeter: UnitProps("millimeter", 3, "mm"),
    Unit.Centimeter: UnitProps("centimeter", 3, "cm"),
    Unit.Meter: UnitProps("meter", 1, "m"),
    Unit.Kilometer: UnitProps("kilometer", 3, "km"),
    Unit.Line: UnitProps("line", 3, "ln"),

    Unit.FootPound: UnitProps('foot-pound', 0, 'ft·lb'),
    Unit.Joule: UnitProps('joule', 0, 'J'),

    Unit.MmHg: UnitProps('mmHg', 0, 'mmHg'),
    Unit.InHg: UnitProps('inHg', 6, 'inHg'),
    Unit.Bar: UnitProps('bar', 2, 'bar'),
    Unit.hPa: UnitProps('hPa', 4, 'hPa'),
    Unit.PSI: UnitProps('psi', 4, 'psi'),

    Unit.Fahrenheit: UnitProps('fahrenheit', 1, '°F'),
    Unit.Celsius: UnitProps('celsius', 1, '°C'),
    Unit.Kelvin: UnitProps('kelvin', 1, '°K'),
    Unit.Rankin: UnitProps('rankin', 1, '°R'),

    Unit.MPS: UnitProps('mps', 0, 'm/s'),
    Unit.KMH: UnitProps('kmh', 1, 'km/h'),
    Unit.FPS: UnitProps('fps', 1, 'ft/s'),
    Unit.MPH: UnitProps('mph', 1, 'mph'),
    Unit.KT: UnitProps('knot', 1, 'kt'),

    Unit.Grain: UnitProps('grain', 1, 'gr'),
    Unit.Ounce: UnitProps('ounce', 1, 'oz'),
    Unit.Gram: UnitProps('gram', 1, 'g'),
    Unit.Pound: UnitProps('pound', 0, 'lb'),
    Unit.Kilogram: UnitProps('kilogram', 3, 'kg'),
    Unit.Newton: UnitProps('newton', 3, 'N'),
}


class AbstractUnit:
    """Abstract class for unit of measure instance definition
    Stores defined unit and value, applies conversions to other prefer_units
    """
    __slots__ = ('_value', '_defined_units')

    def __init__(self, value: [float, int], units: Unit):
        """
        :param units: unit as Unit enum
        :param value: numeric value of the unit
        """
        self._value: float = self.to_raw(value, units)
        self._defined_units: Unit = units

    def __str__(self) -> str:
        """
        :return: readable unit value
        """
        units = self._defined_units
        props = UnitPropsDict[units]
        v = self.from_raw(self._value, units)
        return f'{round(v, props.accuracy)}{props.symbol}'

    def __repr__(self) -> str:
        """
        :return: instance as readable view
        """
        return f'<{self.__class__.__name__}: {self << self.units} ({round(self._value, 4)})>'

    def __float__(self):
        return float(self._value)

    def __eq__(self, other):
        return float(self) == other

    def __lt__(self, other):
        return float(self) < other

    def __gt__(self, other):
        return float(self) > other

    def __le__(self, other):
        return float(self) <= other

    def __ge__(self, other):
        return float(self) >= other

    def __lshift__(self, other: Unit):
        return self.convert(other)

    def __rshift__(self, other: Unit):
        return self.get_in(other)

    def __rlshift__(self, other: Unit):
        return self.convert(other)

    def _unit_support_error(self, value: float, units: Unit):
        """Validates the prefer_units
        :param value: value of the instance
        :param units: Unit enum type
        :return: value in specified prefer_units
        """
        if not isinstance(units, Unit):
            err_msg = f"Type expected: {Unit}, {type(Unit).__name__} " \
                      f"found: {type(units).__name__} ({value})"
            raise TypeError(err_msg)
        if units not in self.__dict__.values():
            raise ValueError(f'{self.__class__.__name__}: unit {units} is not supported')
        return 0

    def to_raw(self, value: float, units: Unit) -> float:
        """Converts value with specified prefer_units to raw value
        :param value: value of the instance
        :param units: Unit enum type
        :return: value in specified prefer_units
        """
        return self._unit_support_error(value, units)

    def from_raw(self, value: float, units: Unit) -> float:
        """Converts raw value to specified prefer_units
        :param value: raw value of the unit
        :param units: Unit enum type
        :return: value in specified prefer_units
        """
        return self._unit_support_error(value, units)

    def convert(self, units: Unit) -> 'AbstractUnit':
        """Returns new unit instance in specified prefer_units
        :param units: Unit enum type
        :return: new unit instance in specified prefer_units
        """
        value = self.get_in(units)
        return self.__class__(value, units)

    def get_in(self, units: Unit) -> float:
        """
        :param units: Unit enum type
        :return: value in specified prefer_units
        """
        return self.from_raw(self._value, units)

    @property
    def units(self) -> Unit:
        """
        :return: defined prefer_units
        """
        return self._defined_units

    @property
    def unit_value(self) -> float:
        """Returns float value in defined prefer_units"""
        return self.get_in(self.units)

    @property
    def raw_value(self) -> float:
        """Raw unit value getter
        :return: raw unit value
        """
        return self._value


class Distance(AbstractUnit):
    """Distance unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Distance.Inch:
            return value
        if units == Distance.Foot:
            result = value * 12
        elif units == Distance.Yard:
            result = value * 36
        elif units == Distance.Mile:
            result = value * 63360
        elif units == Distance.NauticalMile:
            result = value * 72913.3858
        elif units == Distance.Line:
            result = value / 10
        elif units == Distance.Millimeter:
            result = value / 25.4
        elif units == Distance.Centimeter:
            result = value / 2.54
        elif units == Distance.Meter:
            result = value / 25.4 * 1000
        elif units == Distance.Kilometer:
            result = value / 25.4 * 1000000
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Distance.Inch:
            return value
        if units == Distance.Foot:
            result = value / 12
        elif units == Distance.Yard:
            result = value / 36
        elif units == Distance.Mile:
            result = value / 63360
        elif units == Distance.NauticalMile:
            result = value / 72913.3858
        elif units == Distance.Line:
            result = value * 10
        elif units == Distance.Millimeter:
            result = value * 25.4
        elif units == Distance.Centimeter:
            result = value * 2.54
        elif units == Distance.Meter:
            result = value * 25.4 / 1000
        elif units == Distance.Kilometer:
            result = value * 25.4 / 1000000
        else:
            return super().from_raw(value, units)
        return result

    Inch = Unit.Inch
    Foot = Unit.Foot
    Yard = Unit.Yard
    Mile = Unit.Mile
    NauticalMile = Unit.NauticalMile
    Millimeter = Unit.Millimeter
    Centimeter = Unit.Centimeter
    Meter = Unit.Meter
    Kilometer = Unit.Kilometer
    Line = Unit.Line


class Pressure(AbstractUnit):
    """Pressure unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Pressure.MmHg:
            return value
        if units == Pressure.InHg:
            result = value * 25.4
        elif units == Pressure.Bar:
            result = value * 750.061683
        elif units == Pressure.hPa:
            result = value * 750.061683 / 1000
        elif units == Pressure.PSI:
            result = value * 51.714924102396
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Pressure.MmHg:
            return value
        if units == Pressure.InHg:
            result = value / 25.4
        elif units == Pressure.Bar:
            result = value / 750.061683
        elif units == Pressure.hPa:
            result = value / 750.061683 * 1000
        elif units == Pressure.PSI:
            result = value / 51.714924102396
        else:
            return super().from_raw(value, units)
        return result

    MmHg = Unit.MmHg
    InHg = Unit.InHg
    Bar = Unit.Bar
    hPa = Unit.hPa
    PSI = Unit.PSI


class Weight(AbstractUnit):
    """Weight unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Weight.Grain:
            return value
        if units == Weight.Gram:
            result = value * 15.4323584
        elif units == Weight.Kilogram:
            result = value * 15432.3584
        elif units == Weight.Newton:
            result = value * 151339.73750336
        elif units == Weight.Pound:
            result = value / 0.000142857143
        elif units == Weight.Ounce:
            result = value * 437.5
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Weight.Grain:
            return value
        if units == Weight.Gram:
            result = value / 15.4323584
        elif units == Weight.Kilogram:
            result = value / 15432.3584
        elif units == Weight.Newton:
            result = value / 151339.73750336
        elif units == Weight.Pound:
            result = value * 0.000142857143
        elif units == Weight.Ounce:
            result = value / 437.5
        else:
            return super().from_raw(value, units)
        return result

    Grain = Unit.Grain
    Ounce = Unit.Ounce
    Gram = Unit.Gram
    Pound = Unit.Pound
    Kilogram = Unit.Kilogram
    Newton = Unit.Newton


class Temperature(AbstractUnit):
    """Temperature unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Temperature.Fahrenheit:
            return value
        if units == Temperature.Rankin:
            result = value - 459.67
        elif units == Temperature.Celsius:
            result = value * 9 / 5 + 32
        elif units == Temperature.Kelvin:
            result = (value - 273.15) * 9 / 5 + 32
        else:
            return super().to_raw(value, units)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Temperature.Fahrenheit:
            return value
        if units == Temperature.Rankin:
            result = value + 459.67
        elif units == Temperature.Celsius:
            result = (value - 32) * 5 / 9
        elif units == Temperature.Kelvin:
            result = (value - 32) * 5 / 9 + 273.15
        else:
            return super().from_raw(value, units)
        return result

    Fahrenheit = Unit.Fahrenheit
    Celsius = Unit.Celsius
    Kelvin = Unit.Kelvin
    Rankin = Unit.Rankin


class Angular(AbstractUnit):
    """Angular unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Angular.Radian:
            return value
        if units == Angular.Degree:
            result = value / 180 * pi
        elif units == Angular.MOA:
            result = value / 180 * pi / 60
        elif units == Angular.Mil:
            result = value / 3200 * pi
        elif units == Angular.MRad:
            result = value / 1000
        elif units == Angular.Thousandth:
            result = value / 3000 * pi
        elif units == Angular.InchesPer100Yd:
            result = atan(value / 3600)
        elif units == Angular.CmPer100M:
            result = atan(value / 10000)
        elif units == Angular.OClock:
            result = value / 6 * pi
        else:
            return super().to_raw(value, units)
        if result > 2 * pi:
            result = result % (2 * pi)
        return result

    def from_raw(self, value: float, units: Unit):
        if units == Angular.Radian:
            return value
        if units == Angular.Degree:
            result = value * 180 / pi
        elif units == Angular.MOA:
            result = value * 180 / pi * 60
        elif units == Angular.Mil:
            result = value * 3200 / pi
        elif units == Angular.MRad:
            result = value * 1000
        elif units == Angular.Thousandth:
            result = value * 3000 / pi
        elif units == Angular.InchesPer100Yd:
            result = tan(value) * 3600
        elif units == Angular.CmPer100M:
            result = tan(value) * 10000
        elif units == Angular.OClock:
            result = value * 6 / pi
        else:
            return super().from_raw(value, units)
        return result

    Radian = Unit.Radian
    Degree = Unit.Degree
    MOA = Unit.MOA
    Mil = Unit.Mil
    MRad = Unit.MRad
    Thousandth = Unit.Thousandth
    InchesPer100Yd = Unit.InchesPer100Yd
    CmPer100M = Unit.CmPer100M
    OClock = Unit.OClock


class Velocity(AbstractUnit):
    """Velocity unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Velocity.MPS:
            return value
        if units == Velocity.KMH:
            return value / 3.6
        if units == Velocity.FPS:
            return value / 3.2808399
        if units == Velocity.MPH:
            return value / 2.23693629
        if units == Velocity.KT:
            return value / 1.94384449
        return super().to_raw(value, units)

    def from_raw(self, value: float, units: Unit):
        if units == Velocity.MPS:
            return value
        if units == Velocity.KMH:
            return value * 3.6
        if units == Velocity.FPS:
            return value * 3.2808399
        if units == Velocity.MPH:
            return value * 2.23693629
        if units == Velocity.KT:
            return value * 1.94384449
        return super().from_raw(value, units)

    MPS = Unit.MPS
    KMH = Unit.KMH
    FPS = Unit.FPS
    MPH = Unit.MPH
    KT = Unit.KT


class Energy(AbstractUnit):
    """Energy unit"""

    def to_raw(self, value: float, units: Unit):
        if units == Energy.FootPound:
            return value
        if units == Energy.Joule:
            return value * 0.737562149277
        return super().to_raw(value, units)

    def from_raw(self, value: float, units: Unit):
        if units == Energy.FootPound:
            return value
        if units == Energy.Joule:
            return value / 0.737562149277
        return super().from_raw(value, units)

    FootPound = Unit.FootPound
    Joule = Unit.Joule


class PreferredUnitsMeta(type):
    """Provide representation method for static dataclasses."""

    def __repr__(cls):
        return '\n'.join(f'{field} = {getattr(cls, field)!r}'
                         for field in getattr(cls, '__dataclass_fields__'))


@dataclass
class PreferredUnits(metaclass=PreferredUnitsMeta):  # pylint: disable=too-many-instance-attributes
    """Default prefer_units for specified measures"""

    angular: Unit = Unit.Degree
    distance: Unit = Unit.Yard
    velocity: Unit = Unit.FPS
    pressure: Unit = Unit.InHg
    temperature: Unit = Unit.Fahrenheit
    diameter: Unit = Unit.Inch
    length: Unit = Unit.Inch
    weight: Unit = Unit.Grain
    adjustment: Unit = Unit.Mil
    drop: Unit = Unit.Inch
    energy: Unit = Unit.FootPound
    ogw: Unit = Unit.Pound
    sight_height: Unit = Unit.Inch
    target_height: Unit = Unit.Inch
    twist: Unit = Unit.Inch

    @dataclass
    class Mixin(ABC):  # pylint: disable=too-few-public-methods
        """
        TODO: move it to Units, use it instead of TypedUnits
        Abstract class to apply auto-conversion values to
        specified prefer_units by type-hints in inherited dataclasses
        """

        def __setattr__(self, key, value):
            """
            converts value to specified prefer_units by type-hints in inherited dataclass
            """

            _fields = self.__getattribute__('__dataclass_fields__')

            if (_field := _fields.get(key)) and value is not None and not isinstance(value, AbstractUnit):

                if units := _field.metadata.get('prefer_units'):

                    if isinstance(units, Unit):
                        value = units(value)
                    elif isinstance(units, str):
                        value = PreferredUnits.__dict__[units](value)
                    else:
                        raise TypeError(f"Unsupported unit or dimension use one of {PreferredUnits}")

            super().__setattr__(key, value)

    @classmethod
    def defaults(self):
        """resets preferred units to defaults"""
        self.angular = Unit.Degree
        self.distance = Unit.Yard
        self.velocity = Unit.FPS
        self.pressure = Unit.InHg
        self.temperature = Unit.Fahrenheit
        self.diameter = Unit.Inch
        self.length = Unit.Inch
        self.weight = Unit.Grain
        self.adjustment = Unit.Mil
        self.drop = Unit.Inch
        self.energy = Unit.FootPound
        self.ogw = Unit.Pound
        self.sight_height = Unit.Inch
        self.target_height = Unit.Inch
        self.twist = Unit.Inch

    @classmethod
    def set(cls, **kwargs):
        """set preferred units from Mapping"""
        for attribute, value in kwargs.items():

            if hasattr(PreferredUnits, attribute):
                if isinstance(value, Unit):
                    setattr(PreferredUnits, attribute, value)
                elif isinstance(value, str):
                    try:
                        setattr(PreferredUnits, attribute, Unit[value])
                    except KeyError:
                        logger.warning(f"{value=} not a member of Unit")
                else:
                    logger.warning(f"type of {value=} have not been converted to a member of Unit")
            else:
                logger.warning(f"{attribute=} not found in preferred_units")



# pylint: disable=redefined-builtin,too-few-public-methods,too-many-arguments
class Dimension(Field):
    """
    Definition of measure units specified field for
    PreferredUnits.Mixin based dataclasses
    """

    def __init__(self, prefer_units: Union[str, Unit], init=True, repr_=True,
                 hash_=None, compare=True, metadata=None):
        if metadata is None:
            metadata = {}
        metadata['prefer_units'] = prefer_units

        major, minor = sys.version_info.major, sys.version_info.minor

        if major >= 3 and minor > 9:
            extra = {'kw_only': MISSING}
        elif major >= 3 and minor == 9:
            extra = {}
        else:
            raise RuntimeError("Unsupported python version")

        super().__init__(default=None, default_factory=MISSING,
                         init=init, repr=repr_,
                         hash=hash_, compare=compare,
                         metadata=metadata, **extra)

    @abstractmethod
    def __rshift__(self, other):
        ...

    @abstractmethod
    def __lshift__(self, other):
        ...



"""LGPL library for small arms ballistic calculations (Python 3.8+)"""

__author__ = "o-murphy"
__copyright__ = (
    "Copyright 2023 Dmytro Yaroshenko (https://github.com/o-murphy)",
    "Copyright 2024 David Bookstaber (https://github.com/dbookstaber)"
)

__credits__ = ["o-murphy", "dbookstaber"]

import os

from .backend import *
from .drag_tables import *
from .drag_model import *
from .interface import *
from .logger import logger
from .trajectory_data import *
from .conditions import *
from .munition import *
from .unit import *

try:
    import tomllib
except ImportError:
    import tomli as tomllib


def _load_config(filepath=None):

    def find_pybc_toml(start_dir=os.getcwd()):
        """
        Search for the pyproject.toml file starting from the specified directory.
        :param start_dir: (str) The directory to start searching from. Default is the current working directory.
        :return: str: The absolute path to the pyproject.toml file if found, otherwise None.
        """
        current_dir = os.path.abspath(start_dir)
        while True:
            # Check if pybc.toml or .pybc.toml exists in the current directory
            pybc_paths = [
                os.path.join(current_dir, '.pybc.toml'),
                os.path.join(current_dir, 'pybc.toml'),
            ]
            for pypc_path in pybc_paths:
                if os.path.exists(pypc_path):
                    return os.path.abspath(pypc_path)

            # Move to the parent directory
            parent_dir = os.path.dirname(current_dir)

            # If we have reached the root directory, stop searching
            if parent_dir == current_dir:
                return None

            current_dir = parent_dir

    if filepath is None:
        if (filepath := find_pybc_toml()) is None:
            find_pybc_toml(os.path.dirname(__file__))

    logger.debug(f"Found {os.path.basename(filepath)} at {os.path.dirname(filepath)}")

    with open(filepath, "rb") as fp:
        _config = tomllib.load(fp)

        if _pybc := _config.get('pybc'):
            if preferred_units := _pybc.get('preferred_units'):
                PreferredUnits.set(**preferred_units)
            else:
                logger.warning("Config has not `pybc.preferred_units` section")

            if calculator := _pybc.get('calculator'):
                if max_calc_step_size := calculator.get('max_calc_step_size'):
                    try:
                        _val = max_calc_step_size.get("value")
                        _units = Unit[max_calc_step_size.get("units")]
                        set_global_max_calc_step_size(_units(_val))
                    except (KeyError, TypeError, ValueError):
                        logger.warning("Wrong max_calc_step_size units or value")

                if use_powder_sensitivity := calculator.get('use_powder_sensitivity'):
                    set_global_use_powder_sensitivity(use_powder_sensitivity)
            else:
                logger.warning("Config has not `pybc.calculator` section")
        else:
            logger.warning("Config has not `pybc` section")


def _basic_config(filename=None,
                  max_calc_step_size: [float, Distance] = None,
                  use_powder_sensitivity: bool = False,
                  preferred_units: dict[str, Unit] = None):

    """
    Method to load preferred units from file or Mapping
    """
    if filename and (preferred_units or max_calc_step_size or use_powder_sensitivity):
        raise ValueError("Can't use preferred_units and config file at same time")
    if not filename and (preferred_units or max_calc_step_size or use_powder_sensitivity):
        if preferred_units:
            PreferredUnits.set(**preferred_units)
        if max_calc_step_size:
            set_global_max_calc_step_size(max_calc_step_size)
        if use_powder_sensitivity:
            set_global_use_powder_sensitivity(use_powder_sensitivity)
    else:
        # trying to load definitions from pybc.toml
        _load_config(filename)


basicConfig = _basic_config

basicConfig()

__all__ = [
    'Calculator',
    'basicConfig',
    'logger',
    'TrajectoryCalc',
    'get_global_max_calc_step_size',
    'get_global_use_powder_sensitivity',
    'set_global_max_calc_step_size',
    'set_global_use_powder_sensitivity',
    'reset_globals',
    'DragModel',
    'DragDataPoint',
    'BCPoint',
    'DragModelMultiBC',
    'TrajectoryData',
    'HitResult',
    'TrajFlag',
    'Atmo',
    'Wind',
    'Shot',
    'Weapon',
    'Ammo',
    'Unit',
    'AbstractUnit',
    'AbstractUnitType',
    'UnitProps',
    'UnitPropsDict',
    'Distance',
    'Velocity',
    'Angular',
    'Temperature',
    'Pressure',
    'Energy',
    'Weight',
    'Dimension',
    'PreferredUnits'
]

__all__ += ["TableG%s" % n for n in (1, 7, 2, 5, 6, 8, 'I', 'S')]



"""Module for Weapon and Ammo properties definitions"""
import math
from dataclasses import dataclass, field

from .drag_model import DragModel
from .unit import Velocity, Temperature, Distance, Angular, PreferredUnits, Dimension

__all__ = ('Weapon', 'Ammo')


@dataclass
class Weapon(PreferredUnits.Mixin):
    """
    :param sight_height: Vertical distance from center of bore line to center of sight line.
    :param twist: Distance for barrel rifling to complete one complete turn.
        Positive value => right-hand twist, negative value => left-hand twist.
    :param zero_elevation: Angle of barrel relative to sight line when sight is set to "zero."
        (Typically computed by ballistic Calculator.)
    """
    sight_height: [float, Distance] = Dimension(prefer_units='sight_height')
    twist: [float, Distance] = Dimension(prefer_units='twist')
    zero_elevation: [float, Angular] = Dimension(prefer_units='angular')

    def __post_init__(self):
        if not self.sight_height:
            self.sight_height = 0
        if not self.twist:
            self.twist = 0
        if not self.zero_elevation:
            self.zero_elevation = 0


@dataclass
class Ammo(PreferredUnits.Mixin):
    """
    :param dm: DragModel for projectile
    :param mv: Muzzle Velocity
    :param powder_temp: Baseline temperature that produces the given mv
    :param temp_modifier: Change in velocity w temperature: % per 15°C.
        Can be computed with .calc_powder_sens().  Only applies if:
            Settings.USE_POWDER_SENSITIVITY = True
    """
    dm: DragModel = field(default=None)
    mv: [float, Velocity] = Dimension(prefer_units='velocity')
    powder_temp: [float, Temperature] = Dimension(prefer_units='temperature')
    temp_modifier: float = field(default=0)

    def __post_init__(self):
        if not self.powder_temp:
            self.powder_temp = Temperature.Celsius(15)

    def calc_powder_sens(self, other_velocity: [float, Velocity],
                         other_temperature: [float, Temperature]) -> float:
        """Calculates velocity correction by temperature change; assigns to self.temp_modifier
        :param other_velocity: other velocity at other_temperature
        :param other_temperature: other temperature
        :return: temperature modifier in terms %v_delta/15°C
        """
        v0 = self.mv >> Velocity.MPS
        t0 = self.powder_temp >> Temperature.Celsius
        v1 = PreferredUnits.velocity(other_velocity) >> Velocity.MPS
        t1 = PreferredUnits.temperature(other_temperature) >> Temperature.Celsius

        v_delta = math.fabs(v0 - v1)
        t_delta = math.fabs(t0 - t1)
        v_lower = v1 if v1 < v0 else v0

        if v_delta == 0 or t_delta == 0:
            raise ValueError(
                "Temperature modifier error, other velocity"
                " and temperature can't be same as default"
            )
        self.temp_modifier = v_delta / t_delta * (15 / v_lower)  # * 100
        return self.temp_modifier

    def get_velocity_for_temp(self, current_temp: [float, Temperature]) -> Velocity:
        """Calculates muzzle velocity at temperature, based on temp_modifier.
        :param current_temp: Temperature of cartridge powder
        :return: Muzzle velocity corrected to current_temp
        """
        v0 = self.mv >> Velocity.MPS
        t0 = self.powder_temp >> Temperature.Celsius
        t1 = PreferredUnits.temperature(current_temp) >> Temperature.Celsius
        t_delta = t1 - t0
        muzzle_velocity = self.temp_modifier / (15 / v0) * t_delta + v0
        return Velocity.MPS(muzzle_velocity)



import unittest
from py_ballisticcalc import *


class TestDangerSpace(unittest.TestCase):

    def setUp(self) -> None:
        self.look_angle = Angular.Degree(0)
        weight, diameter = 168, 0.308
        length = Distance.Inch(1.282)
        dm = DragModel(0.223, TableG7, weight, diameter, length)
        ammo = Ammo(dm, Velocity.FPS(2750), Temperature.Celsius(15))
        ammo.calc_powder_sens(2723, 0)
        current_winds = [Wind(2, 90)]
        shot = Shot(weapon=Weapon(), ammo=ammo, winds=current_winds)
        calc = Calculator()
        calc.set_weapon_zero(shot, Distance.Foot(300))
        self.shot_result = calc.fire(shot, trajectory_range=Distance.Yard(1000), trajectory_step=Distance.Yard(100), extra_data=True)

    def test_danger_space(self):
        danger_space = self.shot_result.danger_space(
            Distance.Yard(500), Distance.Meter(1.5), self.look_angle
        )

        self.assertAlmostEqual(
            round(danger_space.begin.distance >> Distance.Yard, Distance.Yard.accuracy), 393.6, 0)
        self.assertAlmostEqual(
            round(danger_space.end.distance >> Distance.Yard, Distance.Yard.accuracy), 579.0, 0)

        danger_space = self.shot_result.danger_space(
            Distance.Yard(500), Distance.Inch(10), self.look_angle
        )

        self.assertAlmostEqual(
            round(danger_space.begin.distance >> Distance.Yard, Distance.Yard.accuracy), 484.5, 0)
        self.assertAlmostEqual(
            round(danger_space.end.distance >> Distance.Yard, Distance.Yard.accuracy), 514.8, 0)



"""Run first to ensure binary is run if selected"""
import unittest
#import pyximport; pyximport.install(language_level=3)

class TestImports(unittest.TestCase):

    def test_bin_import(self):
        try:
            import py_ballisticcalc_exts
        except ImportError as err:
            print(err)
            py_ballisticcalc_exts = None



import unittest
from py_ballisticcalc import Atmo
from py_ballisticcalc.unit import *

class TestAtmosphere(unittest.TestCase):
    """Unittests of the Atmosphere class"""

    def setUp(self):
        self.standard = Atmo.standard()
        self.highICAO = Atmo.standard(Distance.Foot(10000))
        self.highISA = Atmo.standard(Distance.Meter(1000))
        self.custom = Atmo(temperature=Temperature.Fahrenheit(30), pressure=Pressure.InHg(31), humidity=0.5)

    def test_standard(self):
        self.assertAlmostEqual(self.standard.temperature >> Temperature.Fahrenheit, 59.0, places=1)
        self.assertAlmostEqual(self.standard.pressure >> Pressure.hPa, 1013.25, places=1)
        self.assertAlmostEqual(self.standard.density_imperial, 0.076474, places=4)

    def test_high(self):
        # Ref https://www.engineeringtoolbox.com/standard-atmosphere-d_604.html
        self.assertAlmostEqual(self.highICAO.temperature >> Temperature.Fahrenheit, 23.36, places=1)
        self.assertAlmostEqual(self.highICAO.density_ratio, 0.7387, places=3)
        # Ref https://www.engineeringtoolbox.com/international-standard-atmosphere-d_985.html
        self.assertAlmostEqual(self.highISA.pressure >> Pressure.hPa, 899, places=0)
        self.assertAlmostEqual(self.highISA.density_ratio, 0.9075, places=4)

    def test_mach(self):
        # Ref https://www.omnicalculator.com/physics/speed-of-sound
        self.assertAlmostEqual(Atmo.machF(59), 1116.15, places=0)
        self.assertAlmostEqual(Atmo.machF(10), 1062.11, places=0)
        self.assertAlmostEqual(Atmo.machF(99), 1158.39, places=0)
        self.assertAlmostEqual(Atmo.machC(-20), 318.94, places=1)
        self.assertAlmostEqual(self.highISA.mach >> Velocity.MPS, 336.4, places=1)


if __name__ == '__main__':
    unittest.main()



"""Implements a point of trajectory class in applicable data types"""
import logging
import math
import typing
from dataclasses import dataclass, field
from enum import Flag
from typing import NamedTuple

from .unit import Angular, Distance, Weight, Velocity, Energy, AbstractUnit, Unit, PreferredUnits
from .conditions import Shot

try:
    import pandas as pd
except ImportError as error:
    logging.warning("Install pandas to convert trajectory to dataframe")
    pd = None

try:
    import matplotlib
    from matplotlib import patches
except ImportError as error:
    logging.warning("Install matplotlib to get results as a plot")
    matplotlib = None

__all__ = ('TrajectoryData', 'HitResult', 'TrajFlag')

PLOT_FONT_HEIGHT = 72
PLOT_FONT_SIZE = 552 / PLOT_FONT_HEIGHT


class TrajFlag(Flag):
    """Flags for marking trajectory row if Zero or Mach crossing
    Also uses to set a filters for a trajectory calculation loop
    """
    NONE = 0
    ZERO_UP = 1
    ZERO_DOWN = 2
    MACH = 4
    RANGE = 8
    DANGER = 16
    ZERO = ZERO_UP | ZERO_DOWN
    ALL = RANGE | ZERO_UP | ZERO_DOWN | MACH | DANGER


class TrajectoryData(NamedTuple):
    """
    Data for one point in ballistic trajectory

    Attributes:
        time (float): bullet flight time
        distance (Distance): x-axis coordinate
        velocity (Velocity): velocity
        mach (float): velocity in Mach prefer_units
        height (Distance): y-axis coordinate
        target_drop (Distance): drop relative to sight-line
        drop_adj (Angular): sight adjustment to zero target_drop at this distance
        windage (Distance):
        windage_adj (Angular):
        look_distance (Distance): sight-line distance = .distance/cosine(look_angle)
        # look_height (Distance): y-coordinate of sight-line = .distance*tan(look_angle)
        angle (Angular): Angle of velocity vector relative to x-axis
        density_factor (float): Ratio of air density here to standard density
        drag (float): Current drag coefficient
        energy (Energy):
        ogw (Weight): optimal game weight
        flag (int): row type
    """

    time: float
    distance: Distance
    velocity: Velocity
    mach: float
    height: Distance
    target_drop: Distance
    drop_adj: Angular
    windage: Distance
    windage_adj: Angular
    look_distance: Distance
    angle: Angular
    density_factor: float
    drag: float
    energy: Energy
    ogw: Weight
    flag: typing.Union[TrajFlag, int]

    def formatted(self) -> tuple:
        """
        :return: matrix of formatted strings for each value of trajectory in default prefer_units
        """

        def _fmt(v: AbstractUnit, u: Unit):
            """simple formatter"""
            return f"{v >> u:.{u.accuracy}f} {u.symbol}"

        return (
            f'{self.time:.3f} s',
            _fmt(self.distance, PreferredUnits.distance),
            _fmt(self.velocity, PreferredUnits.velocity),
            f'{self.mach:.2f} mach',
            _fmt(self.height, PreferredUnits.drop),
            _fmt(self.target_drop, PreferredUnits.drop),
            _fmt(self.drop_adj, PreferredUnits.adjustment),
            _fmt(self.windage, PreferredUnits.drop),
            _fmt(self.windage_adj, PreferredUnits.adjustment),
            _fmt(self.look_distance, PreferredUnits.distance),
            _fmt(self.angle, PreferredUnits.angular),
            f'{self.density_factor:.3e}',
            f'{self.drag:.3f}',
            _fmt(self.energy, PreferredUnits.energy),
            _fmt(self.ogw, PreferredUnits.ogw),

            self.flag
        )

    def in_def_units(self) -> tuple:
        """
        :return: matrix of floats of the trajectory in default prefer_units
        """
        return (
            self.time,
            self.distance >> PreferredUnits.distance,
            self.velocity >> PreferredUnits.velocity,
            self.mach,
            self.height >> PreferredUnits.drop,
            self.target_drop >> PreferredUnits.drop,
            self.drop_adj >> PreferredUnits.adjustment,
            self.windage >> PreferredUnits.drop,
            self.windage_adj >> PreferredUnits.adjustment,
            self.look_distance >> PreferredUnits.distance,
            self.angle >> PreferredUnits.angular,
            self.density_factor,
            self.drag,
            self.energy >> PreferredUnits.energy,
            self.ogw >> PreferredUnits.ogw,
            TrajFlag(self.flag)
        )


class DangerSpace(NamedTuple):
    """Stores the danger space data for distance specified"""
    at_range: TrajectoryData
    target_height: Distance
    begin: TrajectoryData
    end: TrajectoryData
    look_angle: Angular

    def __str__(self) -> str:
        return f'Danger space at {self.at_range.distance << PreferredUnits.distance} ' \
            + f'for {self.target_height << PreferredUnits.drop} tall target ' \
            + (f'at {self.look_angle << Angular.Degree} look-angle ' if self.look_angle != 0 else '') \
            + f'ranges from {self.begin.distance << PreferredUnits.distance} ' \
            + f'to {self.end.distance << PreferredUnits.distance}'

    def overlay(self, ax: 'Axes', label: str = None):
        """Highlights danger-space region on plot"""
        if matplotlib is None:
            raise ImportError("Install matplotlib to get results as a plot")

        cosine = math.cos(self.look_angle >> Angular.Radian)
        begin_dist = (self.begin.distance >> PreferredUnits.distance) * cosine
        begin_drop = (self.begin.height >> PreferredUnits.drop) * cosine
        end_dist = (self.end.distance >> PreferredUnits.distance) * cosine
        end_drop = (self.end.height >> PreferredUnits.drop) * cosine
        range_dist = (self.at_range.distance >> PreferredUnits.distance) * cosine
        range_drop = (self.at_range.height >> PreferredUnits.drop) * cosine
        h = self.target_height >> PreferredUnits.drop

        # Target position and height:
        ax.plot((range_dist, range_dist), (range_drop + h / 2, range_drop - h / 2),
                color='r', linestyle=':')
        # Shaded danger-space region:
        vertices = (
            (begin_dist, begin_drop), (end_dist, begin_drop),
            (end_dist, end_drop), (begin_dist, end_drop)
        )
        polygon = patches.Polygon(vertices, closed=True,
                                  edgecolor='none', facecolor='r', alpha=0.3)
        ax.add_patch(polygon)
        if label is None:  # Add default label
            label = f"Danger space\nat {self.at_range.distance << PreferredUnits.distance}"
        if label != '':
            ax.text(begin_dist + (end_dist - begin_dist) / 2, end_drop, label,
                    linespacing=1.2, fontsize=PLOT_FONT_SIZE, ha='center', va='top')


@dataclass(frozen=True)
class HitResult:
    """Results of the shot"""
    shot: Shot
    trajectory: list[TrajectoryData] = field(repr=False)
    extra: bool = False

    def __iter__(self):
        yield from self.trajectory

    def __getitem__(self, item):
        return self.trajectory[item]

    def __check_extra__(self):
        if not self.extra:
            raise AttributeError(
                f"{object.__repr__(self)} has no extra data. "
                f"Use Calculator.fire(..., extra_data=True)"
            )

    def zeros(self) -> list[TrajectoryData]:
        """:return: zero crossing points"""
        self.__check_extra__()
        data = [row for row in self.trajectory if row.flag & TrajFlag.ZERO.value]
        if len(data) < 1:
            raise ArithmeticError("Can't find zero crossing points")
        return data

    def index_at_distance(self, d: Distance) -> int:
        """
        :param d: Distance for which we want Trajectory Data
        :return: Index of first trajectory row with .distance >= d; otherwise -1
        """
        # Get index of first trajectory point with distance >= at_range
        return next((i for i in range(len(self.trajectory))
                     if self.trajectory[i].distance >= d), -1)

    def get_at_distance(self, d: Distance) -> TrajectoryData:
        """
        :param d: Distance for which we want Trajectory Data
        :return: First trajectory row with .distance >= d
        """

        if i := self.index_at_distance(d) < 0:
            raise ArithmeticError(
                f"Calculated trajectory doesn't reach requested distance {d}"
            )
        return self.trajectory[i]

    def danger_space(self,
                     at_range: [float, Distance],
                     target_height: [float, Distance],
                     look_angle: [float, Angular] = None
                     ) -> DangerSpace:
        """
        Assume that the trajectory hits the center of a target at any distance.
        Now we want to know how much ranging error we can tolerate if the critical region 
        of the target has height *h*.  I.e., we want to know how far forward and backward
        along the line of sight we can move a target such that the trajectory is still 
        within *h*/2 of the original drop.

        :param at_range: Danger space is calculated for a target centered at this sight distance
        :param target_height: Target height (*h*) determines danger space
        :param look_angle: Ranging errors occur along the look angle to the target
        """
        self.__check_extra__()

        at_range = PreferredUnits.distance(at_range)
        target_height = PreferredUnits.distance(target_height)
        target_height_half = target_height.raw_value / 2.0
        if look_angle is None:
            look_angle = self.shot.look_angle
        else:
            look_angle = PreferredUnits.angular(look_angle)

        # Get index of first trajectory point with distance >= at_range
        if (index := self.index_at_distance(at_range)) < 0:
            raise ArithmeticError(
                f"Calculated trajectory doesn't reach requested distance {at_range}"
            )

        def find_begin_danger(row_num: int) -> TrajectoryData:
            """
            Beginning of danger space is last .distance' < .distance where
                (.drop' - target_center) >= target_height/2
            :param row_num: Index of the trajectory point for which we are calculating danger space
            :return: Distance marking beginning of danger space
            """
            center_row = self.trajectory[row_num]
            for prime_row in reversed(self.trajectory[:row_num]):
                if (prime_row.target_drop.raw_value - center_row.target_drop.raw_value) >= target_height_half:
                    return prime_row
            return self.trajectory[0]

        def find_end_danger(row_num: int) -> TrajectoryData:
            """
            End of danger space is first .distance' > .distance where
                (target_center - .drop') >= target_height/2
            :param row_num: Index of the trajectory point for which we are calculating danger space
            :return: Distance marking end of danger space
            """
            center_row = self.trajectory[row_num]
            for prime_row in self.trajectory[row_num + 1:]:
                if (center_row.target_drop.raw_value - prime_row.target_drop.raw_value) >= target_height_half:
                    return prime_row
            return self.trajectory[-1]

        return DangerSpace(self.trajectory[index],
                           target_height,
                           find_begin_danger(index),
                           find_end_danger(index),
                           look_angle)

    def dataframe(self, formatted: bool = False) -> 'DataFrame':
        """
        :param formatted: False for values as floats; True for strings with prefer_units
        :return: the trajectory table as a DataFrame
        """
        if pd is None:
            raise ImportError("Install pandas to get trajectory as dataframe")
        col_names = list(TrajectoryData._fields)
        if formatted:
            trajectory = [p.formatted() for p in self]
        else:
            trajectory = [p.in_def_units() for p in self]
        return pd.DataFrame(trajectory, columns=col_names)

    def plot(self, look_angle: Angular = None) -> 'Axes':
        """:return: graph of the trajectory"""
        if look_angle is None:
            look_angle = self.shot.look_angle

        if matplotlib is None:
            raise ImportError("Install matplotlib to plot results")
        if not self.extra:
            logging.warning("HitResult.plot: To show extended data"
                            "Use Calculator.fire(..., extra_data=True)")
        font_size = PLOT_FONT_SIZE
        df = self.dataframe()
        ax = df.plot(x='distance', y=['height'], ylabel=PreferredUnits.drop.symbol)
        max_range = self.trajectory[-1].distance >> PreferredUnits.distance

        for p in self.trajectory:
            if TrajFlag(p.flag) & TrajFlag.ZERO:
                ax.plot([p.distance >> PreferredUnits.distance, p.distance >> PreferredUnits.distance],
                        [df['height'].min(), p.height >> PreferredUnits.drop], linestyle=':')
                ax.text((p.distance >> PreferredUnits.distance) + max_range / 100, df['height'].min(),
                        f"{(TrajFlag(p.flag) & TrajFlag.ZERO).name}",
                        fontsize=font_size, rotation=90)
            if TrajFlag(p.flag) & TrajFlag.MACH:
                ax.plot([p.distance >> PreferredUnits.distance, p.distance >> PreferredUnits.distance],
                        [df['height'].min(), p.height >> PreferredUnits.drop], linestyle=':')
                ax.text((p.distance >> PreferredUnits.distance) + max_range / 100, df['height'].min(),
                        "Mach 1", fontsize=font_size, rotation=90)

        max_range_in_drop_units = self.trajectory[-1].distance >> PreferredUnits.drop
        # Sight line
        x_sight = [0, df.distance.max()]
        y_sight = [0, max_range_in_drop_units * math.tan(look_angle >> Angular.Radian)]
        ax.plot(x_sight, y_sight, linestyle='--', color=[.3, 0, .3, .5])
        # Barrel pointing line
        x_bbl = [0, df.distance.max()]
        y_bbl = [-(self.shot.weapon.sight_height >> PreferredUnits.drop),
                 max_range_in_drop_units * math.tan(self.trajectory[0].angle >> Angular.Radian)
                 - (self.shot.weapon.sight_height >> PreferredUnits.drop)]
        ax.plot(x_bbl, y_bbl, linestyle=':', color=[0, 0, 0, .5])
        # Line labels
        sight_above_bbl = y_sight[1] > y_bbl[1]
        angle = math.degrees(math.atan((y_sight[1] - y_sight[0]) / (x_sight[1] - x_sight[0])))
        ax.text(x_sight[1], y_sight[1], "Sight line", linespacing=1.2,
                rotation=angle, rotation_mode='anchor', transform_rotates_text=True,
                fontsize=font_size, color=[.3, 0, .3, 1], ha='right',
                va='bottom' if sight_above_bbl else 'top')
        angle = math.degrees(math.atan((y_bbl[1] - y_bbl[0]) / (x_bbl[1] - x_bbl[0])))
        ax.text(x_bbl[1], y_bbl[1], "Barrel pointing", linespacing=1.2,
                rotation=angle, rotation_mode='anchor', transform_rotates_text=True,
                fontsize=font_size, color='k', ha='right',
                va='top' if sight_above_bbl else 'bottom')
        # Plot velocity (on secondary axis)
        df.plot(x='distance', xlabel=PreferredUnits.distance.symbol,
                y=['velocity'], ylabel=PreferredUnits.velocity.symbol,
                secondary_y=True, color=[0, .3, 0, .5],
                ylim=[0, df['velocity'].max()], ax=ax)
        # Let secondary shine through
        ax.set_zorder(1)
        ax.set_facecolor([0, 0, 0, 0])

        return ax



import os
from unittest import TestCase
from py_ballisticcalc import (basicConfig, PreferredUnits, Unit,
                              get_global_max_calc_step_size, reset_globals)

ASSETS_DIR = os.path.join(
    os.path.dirname(
        os.path.dirname(__file__)
    ), 'assets')


class TestConfigLoader(TestCase):

    def test_preferred_units_load(self):
        with self.subTest("env"):
            basicConfig()
            self.assertEqual(PreferredUnits.distance, Unit.Yard)

        with self.subTest("manual"):
            basicConfig(max_calc_step_size=Unit.Meter(0.3), preferred_units={
                'distance': Unit.Meter
            })
            self.assertEqual(get_global_max_calc_step_size().units, Unit.Meter)
            self.assertEqual(PreferredUnits.distance, Unit.Meter)

        with self.subTest("imperial"):
            basicConfig(os.path.join(ASSETS_DIR, ".pybc-imperial.toml"))
            self.assertEqual(PreferredUnits.distance, Unit.Foot)

        with self.subTest("mixed"):
            basicConfig(os.path.join(ASSETS_DIR, ".pybc-metrics.toml"))
            self.assertEqual(PreferredUnits.distance, Unit.Meter)

        with self.subTest("mixed"):
            basicConfig(os.path.join(ASSETS_DIR, ".pybc-mixed.toml"))
            self.assertEqual(PreferredUnits.velocity, Unit.MPS)

        basicConfig()
        reset_globals()
        PreferredUnits.defaults()
